#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
    Functions for training (semi-supervised) hierarchical NMF model via Neural NMF.

    Examples
    --------
    >>> net = Neural_NMF([m, 9], none)                            #declares object from Neural NMF module
    >>> train(net, X, epoch=200, lr=1000)                            
    epoch =  10 
     tensor(228.1642, dtype=torch.float64)
    epoch =  20 
     tensor(228.0765, dtype=torch.float64)
    epoch =  30
    ...
'''

from NeuralNMF import Neural_NMF, Recon_Loss_Func, Energy_Loss_Func, Fro_Norm
from NeuralNMF import Writer
import torch
import numpy as np
from NeuralNMF import LsqNonneg
import torch.nn as nn
from matplotlib import pyplot as plt
from tqdm import trange
from torch.autograd import Variable


def train(net, X, loss_func="Recon Loss", supervised=False, label=None, L=None, epoch=10, lr=1e-3, weight_decay=1, class_iters=1, decay_epoch=1, initialize_support = False, optimizer ="gd", verbose=True,verbose_epoch=1, full_history=False):

    """
    Training the Neural_NMF with projected gradient descent.
    
    Parameters
    ----------
    net: Pytorch Neural NMF module object
        The Neural NMF object to be trained. 
    X: PyTorch Tensor
        The data matrix input into the Neural NMF network (m x n).
    loss_func: string or Pytorch Module, optional
        The loss function. If the input is a string, uses one of the default loss functions 'Recon Loss' 
        or 'Energy Loss' (default is 'Recon Loss'). If the input is a Pytorch Module, uses that module, 
        allowing for custom loss functions.
    supervised: bool, optional
        Indicator for supervision (default is False).
    label: PyTorch tensor, optional
        The classification (label) matrix for supervised model.  If the classification_type is 'L2',
        this matrix is a one-hot encoding matrix of size c x n.  If the classification_type is
        'CrossEntropy', this matrix is of size n with elements in [0,c-1] (default is None).
    L: PyTorch tensor, optional
        The label indicator matrix for semi-supervised model that indicates if labels are known 
        for n data points, of size c x n with columns of all ones or all zeros to indicate if label
        for that data point is known (default is None).
    epoch: int_, optional
        Number of epochs in training procedure (default 10).
    lr: float_, optional
        The learning rate (default 1e-3).
    weight_decay: float, optional
        The weight decay parameter, doing lr = lr*weight_decay every decay_epoch epochs (default 1).
    class_iters: int_, optional
        Number of gradient steps to take on classification term each epoch (default 1).
    decay_epoch: int_, optional
        Number of epochs to take before decaying the learning rates (default 1).
    initialize_support: bool, optional
        Indicator for whether to initialize the estimated support for the least squares solution
        based on the previous support, when using the fnnls method
    optimizer: string, optional
        Optimizer to use to optimize Neural NMF model, either 'gd' for gradient descent of 'adam' for adam optimizer
    verbose: bool, optional
        Indicator for whether to print the loss every verbose_epoch epochs (default True).
    verbose_epoch: int_, optional
        Number of epochs to take before printing the loss (default 1).
    full_history: bool, optional
        Indicator for whether to save all information from every epoch (default False).

    Returns
    -------
    history: Writer object or pair of lists of PyTorch tensors
        If Writer object (full_history=True), the history of the loss, A and S matrices, and A gradients 
        (if supervised training, also stores weights and gradients from the linear layer).
        If pair of lists of PyTorch tensors (full_history=False), stores the final A and S matrices and 
        A gradients from after last epoch.

    """
    if len(X.shape) != 2:
            raise ValueError("Expected a two-dimensional Tensor, but X is of shape {}".format(X.shape))
            
    if optimizer != "gd" and optimizer != "adam":
        raise ValueError("Expected 'gd' or 'adam' for optimizer, but optimizer parameter is {}".format(optimizer))

    if supervised == False:

        history = train_unsupervised(net, X, loss_func=loss_func, epoch = epoch, lr = lr, weight_decay = weight_decay, decay_epoch=decay_epoch, initialize_support=initialize_support, optimizer=optimizer, verbose=verbose, verbose_epoch=verbose_epoch, full_history = full_history)

    else:
        history = train_supervised(net, X, label=label, L=L, loss_func=loss_func, epoch = 10, lr = lr, weight_decay = weight_decay, class_iters=class_iters, initialize_support=initialize_support, decay_epoch=decay_epoch, optimizer=optimizer, verbose=verbose, verbose_epoch=verbose_epoch, full_history = full_history)

    return history



def train_unsupervised(net, X, loss_func="Recon Loss", epoch = 10, lr = 1e-3, weight_decay = 1, decay_epoch=1, initialize_support = False, optimizer="gd", verbose=True, verbose_epoch=1, full_history=False):
    """
    Training the unsupervised Neural_NMF with projected gradient descent.
    
    Parameters
    ----------
    net: Pytorch Neural NMF module object
        The Neural NMF object to be trained (since unsupervised, net.c = None). 
    X: PyTorch Tensor
        The data matrix input into the Neural NMF network (m x n).
    loss_func: string or Pytorch Module, optional
        The loss function. If the input is a string, uses one of the default loss functions 'Recon Loss' 
        or 'Energy Loss' (default is 'Recon Loss'). If the input is a Pytorch Module, uses that module, 
        allowing for custom loss functions. For unsupervised training, any custom loss function should 
        have lambd = None or 0.
    epoch: int_, optional
        Number of epochs in training procedure (default 10).
    lr: float_, optional
        The learning rate for the NMF layers (default 1e-3).
    weight_decay: float, optional
        The weight decay parameter, doing lr = lr*weight_decay every decay_epoch epochs (default 1).
    decay_epoch: int_, optional
        Number of epochs to take before decaying the learning rates (default 1).
    initialize_support: bool, optional
        Indicator for whether to initialize the estimated support for the least squares solution
        based on the previous support, when using the fnnls method
    optimizer: string, optional
        Optimizer to use to optimize Neural NMF model, either 'gd' for gradient descent of 'adam' for adam optimizer
    verbose: bool, optional
        Indicator for whether to print the loss every verbose_epoch epochs (default True).
    verbose_epoch: int_, optional
        Number of epochs to take before printing the loss (default 1).
    full_history: bool, optional
        Indicator for whether to save all information from every epoch (default False).

    Returns
    -------
    history: Writer object, optional
        If full_history=True, stores the history of the loss, A and S matrices, 
        and A gradients at each epoch.
    A_lst: list, optional
        If full_history=False, stores a list of the A PyTorch tensors, A_0, A_1, ... A_L.
    S_lst: list, optional
        If full_history=False, stores a list of the S PyTorch tensors, S_0, S_1, ... S_L.
        
    """
    
    if optimizer != "gd" and optimizer != "adam":
        raise ValueError("Expected 'gd' of 'adam' for optimizer, but optimizer parameter is {}".format(optimizer))

    loss_functions = {"Recon Loss": Recon_Loss_Func(), "Energy Loss": Energy_Loss_Func()}

    A_lst = []
    history = Writer() # creating a Writer object to record the history for the training process

    configs = [{} for i in range(net.depth-1)]

    for config in configs:
        config['learning_rate'] = 10e4
        config['beta1'] = 0.9
        config['beta2'] = 0.99
        config['epsilon'] = 1e-8
        config['t'] = 0

    S_lst = None
    
    if(verbose):
        f = trange
    else:
        f = np.arange
            
    for i in f(epoch):
        net.zero_grad()
        if S_lst != None and initialize_support:
            S_lst = net(X, [s.detach() for s in S_lst])
        else:
            S_lst = net(X)
        loss = None
        if type(loss_func) == str:
            loss = loss_functions[loss_func](net, X, S_lst)
        else:
            loss = loss_func(net, X, S_lst)
        loss.backward()
        history.add_scalar('loss', loss.data)
        for l in range(net.depth - 1):

            if epoch == 0:
                for config in configs:
                    config['v'] = torch.zeros_like(As[l].data)
                    config['a'] = torch.zeros_like(As[l].data)

            A = net.lsqnonneglst[l].A
            # record history
            if full_history:
                history.add_tensor('A'+str(l+1), A.data.clone())
                history.add_tensor('grad_A'+str(l+1), A.grad.data.clone())
                history.add_tensor('S' + str(l+1), S_lst[l].data.clone())

            if not full_history and i == epoch-1:
               A_lst.append(A)

            if(optimizer == "adam"):
                A.data, configs[l] = adam(A.data, A.grad.data, configs[l])
            else:
                #projection gradient descent
                A.data = A.data.sub_(lr*A.grad.data)
        
        #if verbose and (i+1)%verbose_epoch == 0:
        #    print('epoch = ', i+1, '\n', loss.data)
        if (i+1)%decay_epoch == 0:
            lr = lr*weight_decay

    if full_history:        
        return history
    else:
        return A_lst, S_lst

def train_supervised(net, X, label, L = None, loss_func="Recon Loss", epoch = 10, lr = 1e-3, weight_decay = 1, class_iters=1, decay_epoch=1, initialize_support = False, optimizer="gd", verbose=True, verbose_epoch=1, full_history=False):

    """
    Training the supervised Neural_NMF with projected gradient descent (PGD). In each epoch, we update the NMF
    layers with one PGD step and the classification layer with one PGD step.

    Parameters
    ----------
    net: Pytorch Neural NMF module object
        The Neural NMF object to be trained. 
    X: PyTorch tensor
        The data matrix input into the Neural NMF network (m x n).
    label: PyTorch tensor
        The classification (label) matrix for supervised model.  If the classification_type is 'L2',
        this matrix is a one-hot encoding matrix of size c x n.  If the classification_type is
        'CrossEntropy', this matrix is of size n with elements in [0,c-1] (default is None).
    L: PyTorch tensor, optional
        The label indicator matrix for semi-supervised model that indicates if labels are known 
        for n data points, of size c x n with columns of all ones or all zeros to indicate if label
        for that data point is known (default is None).
    loss_func: string or Pytorch Module, optional
        The loss function. If the input is a string, uses one of the default loss functions 'Recon Loss' 
        or 'Energy Loss' (default is 'Recon Loss'). If the input is a Pytorch Module, uses that module, 
        allowing for custom loss functions. 
    epoch: int_, optional
        Number of epochs in training procedure (default 10).
    lr: float_, optional
        The learning rate.
    weight_decay: float, optional
        The weight decay parameter, doing lr = lr*weight_decay every decay_epoch epochs (default 1).
    class_iters: int_, optional
        Number of PGD updates to make to classification layer each epoch (default 1).
    decay_epoch: int_, optional
        Number of epochs to take before decaying the learning rates (default 1).
    initialize_support: bool, optional
        Indicator for whether to initialize the estimated support for the least squares solution
        based on the previous support, when using the fnnls method
    optimizer: string, optional
        Optimizer to use to optimize Neural NMF model, either 'gd' for gradient descent or 'adam' for adam optimizer
    verbose: bool, optional
        Indicator for whether to print the loss every verbose_epoch epochs (default True).
    verbose_epoch: int_, optional
        Number of epochs to take before printing the loss (default 1).
    full_history: bool, optional
        Indicator for whether to save all information from every epoch (default False).

    Returns
    -------
    history: Writer object, optional
        If full_history=True, stores the history of the loss, A and S matrices, 
        and A gradients at each epoch.
    A_lst: list, optional
        If full_history=False, stores a list of the A PyTorch tensors, A_0, A_1, ... A_L.
    S_lst: list, optional
        If full_history=False, stores a list of the S PyTorch tensors, S_0, S_1, ... S_L.

    """
    
    if len(lr) == 2:
        lr_nmf = lr[0]
        lr_class = lr[1]
    else:
        lr_nmf = lr
        lr_class = lr
    
    if optimizer != "gd" and optimizer != "adam":
        raise ValueError("Expected 'gd' of 'adam' for optimizer, but optimizer parameter is {}".format(optimizer))

    loss_functions = {"Recon Loss": Recon_Loss_Func(), "Energy Loss": Energy_Loss_Func()}

    A_lst = []
    history = Writer() # creating a Writer object to record the history for the training process

    if(optimizer == "adam"):
        
        configs = [{} for i in range(net.depth-1)]
        config_w = {}

        for config in configs:
            config['learning_rate'] = lr_nmf
            config['beta1'] = 0.9
            config['beta2'] = 0.99
            config['epsilon'] = 1e-8
            config['t'] = 0

    S_lst = None
    
    
    Y = torch.zeros((net.depth_info[-1],label.shape[0]), dtype=torch.double)
    r = np.arange(X.shape[1])
    Y[[label,r]] = 1
    Y = Y * L
    
    if(verbose):
        f = trange
    else:
        f = np.arange
            
    for i in f(epoch):
        # doing gradient update for NMF layer
        net.zero_grad()
        if S_lst != None and initialize_support:
            S_lst, pred = net(X, Y=Y, L=L, last_S_lst=[s.detach() for s in S_lst])
        else:
            S_lst, pred = net(X, Y=Y, L=L)

        loss = None
        
        if type(loss_func) == str:
            loss_nmf = loss_functions[loss_func](net, X, S_lst)
        else:
            loss_nmf = loss_func(net, X, S_lst)
         
        if type(loss_func) == str:
                loss = loss_functions[loss_func](net, X, S_lst,pred,label,L)
        else:
            loss = loss_func(net, X, S_lst,pred,label,L)
           
                
        loss.backward()
        history.add_scalar('loss_nmf',loss_nmf.data)
        history.add_scalar('loss_classification',loss.data)
        for l in range(net.depth - 1):

            A = net.lsqnonneglst[l].A
            # record history
            if full_history:
                history.add_tensor('A'+str(l+1), A.data.clone())
                history.add_tensor('grad_A'+str(l+1), A.grad.data.clone())
                history.add_tensor('S' + str(l+1), S_lst[l].data.clone())

            if not full_history and i == epoch-1:
                A_lst.append(A)

            if(optimizer == "adam"):
                A.data, configs[l] = adam(A.data, A.grad.data, configs[l])
            else:
                A.data = A.data.sub_(lr_nmf*A.grad.data)

            A.data = A.data.clamp(min = 0)
            
        # doing gradient update for classification layer
        for iter_classifier in range(30):
            net.zero_grad()
            if S_lst != None and initialize_support:
                S_lst, pred = net(X, Y=Y, L=L, last_S_lst=[s.detach() for s in S_lst])
            else:
                S_lst, pred = net(X, Y=Y, L=L)

            loss = None
        
            if type(loss_func) == str:
                loss_nmf = loss_functions[loss_func](net, X, S_lst)
            else:
                loss_nmf = loss_func(net, X, S_lst)
         
            if type(loss_func) == str:
                    loss = loss_functions[loss_func](net, X, S_lst,pred,label,L)
            else:
                loss = loss_func(net, X, S_lst,pred,label,L)
            
            loss.backward()
            S_lst[0].detach()
            history.add_scalar('loss',loss.data)
            weight = net.linear.weight
            weight.data = weight.data.sub_(lr_class*weight.grad.data)
            history.add_tensor('weight', weight.data.clone())
            history.add_tensor('grad_weight', weight.grad.data.clone())
            
        
        #if(verbose) and (i+1)%verbose_epoch == 0:
        #    print('epoch = ', i+1, '\n', loss.data)
        if (i+1)%decay_epoch == 0:
            lr_nmf = lr_nmf*weight_decay
            lr_class = lr_class*weight_decay

    if full_history:        
        return history
    else:
        return A_lst, S_lst


def adam(w, dw, config=None):
    """
    Uses the Adam update rule, which incorporates moving averages of both the
    gradient and its square and a bias correction term.

    config format:
    - learning_rate: Scalar learning rate.
    - beta1: Decay rate for moving average of first moment of gradient.
    - beta2: Decay rate for moving average of second moment of gradient.
    - epsilon: Small scalar used for smoothing to avoid dividing by zero.
    - m: Moving average of gradient.
    - v: Moving average of squared gradient.
    - t: Iteration number.
    """
    if config is None: config = {}
    config.setdefault('learning_rate', 1e-3)
    config.setdefault('beta1', 0.9)
    config.setdefault('beta2', 0.999)
    config.setdefault('epsilon', 1e-8)
    config.setdefault('v', torch.zeros_like(w))
    config.setdefault('a', torch.zeros_like(w))
    config.setdefault('t', 0)
    
    next_w = None

    v = config['v']
    beta1 = config['beta1']
    beta2 = config['beta2']
    rate = config['learning_rate']
    a = config['a']
    e = config['epsilon']
    t = config['t'] + 1

    nu = 1e-8
    v = beta1 * v + (1 - beta1) * dw

    a = beta2 * a + (1 - beta2) * dw * dw 

    v_c  = v * 1 / (1-beta1**t)
    a_c = a * 1 / (1-beta2**t)


    next_w = w - rate * v_c / (np.sqrt(a_c) + e)


    config['v'] = v
    config['a'] = a
    config['t']  = t
    
    return next_w, config




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
    Defines class and functions for Nonnegative Least-Squares layers in hierarchical NMF
    model.
'''

import numpy as np
import torch
import torch.nn as nn
from torch.autograd import Variable
from torch.autograd.function import once_differentiable
from scipy.optimize import nnls

from fnnls import fnnls

class LsqNonnegF(torch.autograd.Function):
    """
    Define the forward and backward process for q(X,A) = argmin_{S >= 0} ||X - AS||_F^2.

    """
    @staticmethod
    def forward(ctx, input, A, last_S=None):
        """
        Runs the forward pass of the nonnegative least squares task q(X,A).

        This solves the problem:
        min_{S >= 0} ||X - A*S||_F^2.
        Output[:,i] = argmin_{s >= 0} ||X[:,i] - A*s||_F^2.

        Parameters
        ----------
        ctx: context object
            Stashes information for the backwards pass.
        input: PyTorch tensor
            The input to the Neural NMF network, X. Input should have size (m,n).
        A: PyTorch tensor
            The A matrix for the current layer of the Neural NMF network, stored in the LsqNonneg class.

        Returns
        -------
        output: PyTorch tensor
            The output of the nonnegative least squares task q(X,A).

        """

        [output, res] = lsqnonneg_tensor_version(A.data, input.data, last_S)
        ctx.save_for_backward(input, A, None)
        ctx.intermediate = output
        return output.t().t()
    
    @staticmethod
    @once_differentiable #without this line then in backprop all the operations should be differentiable
    def backward(ctx, grad_output):
        """
        Runs the backwards pass of the nonnegative least squares task q(X,A). Computes the gradients of 
        q(X,A) with respect to X and A

        Parameters
        ----------
        ctx: context object
            Contains information stashed by the forward pass.
        grad_output: PyTorch tensor
            The gradient of S = q(X,A) passed on from the last layer

        Returns
        -------
        grad_input: PyTorch tensor
            The gradient of the nonnegative least squares task q(X,A) with respect to X.
        grad_A: PyTorch tensor
            The gradient of the nonnegative least squares task q(X,A) with respect to A.

        """

        input, A, temp = ctx.saved_tensors
        grad_input = grad_A = None
        output = ctx.intermediate
        if ctx.needs_input_grad[0]:
            grad_input = calc_grad_X(grad_output, A.data, output)# calculate gradient with respect to X
        if ctx.needs_input_grad[1]:
            grad_A = calc_grad_A(grad_output, A.data, output, input.data) # calculate gradient with respect to A
        return grad_input, grad_A, None




def lsqnonneg_tensor_version(A, X, last_S=None):
    """
    Calculates the nonnegative least squares solution q(X,A).

    Computes the following for each column of the output:
    output[:,i] = argmin_{s >= 0} ||X[:,i] - A*s||_F^2. 

    Parameters
    ---------
    A: PyTorch tensor
        The A matrix used to compute q(X,A).
    X: PyTorch tensor
        The X matrix used to compute q(X,A).

    Returns
    -------
    S: PyTorch tensor
        The S matrix, S = q(X,A).

    """
    A = A.numpy() # Transforming to numpy array size(m,k)
    X = X.numpy() # size(m,n)
    m = X.shape[0]
    n = X.shape[1]
    k = A.shape[1]
    S = np.zeros([k,n])
    res_total = 0
    
    #print([np.max(A),np.max(X)])
    for i in range(n):
        x = X[:,i]

        if last_S != None:
            P_initial = np.asarray([j for j in range(k) if last_S[j,i] > 0], dtype=int)
        else:
            P_initial = np.zeros(0, dtype=int)

        [s, res] = nnls(A, x)
        #[s, res] = fnnls(A, x)
        #[s, res] = fnnls(A, x, P_initial=P_initial)

        res_total += res
        S[:,i] = s

    S = torch.from_numpy(S).double() # Transforming to torch Tensor
    return S, res_total

def calc_grad_X(grad_S, A, S):
    """
    Calculates the gradient of q(X,A) with respect to X.

    Parameters
    ---------
    grad_S: PyTorch tensor
        The gradient of S = q(X,A) passed on from the last layer.
    A: PyTorch tensor
        The A matrix used to compute q(X,A).
    S: PyTorch tensor
        The output S = q(X,A).

    Returns
    -------
    grad_X: PyTorch tensor
        The gradient of q(X,A) with respsect to X.

    """
    A_np = A.numpy()
    S_np = S.numpy()
    grad_S_np = grad_S.numpy()
    m = A.shape[0]
    k = A.shape[1]
    n = S.shape[1]
    grad_X = np.zeros([m,n])
    for i in range(n):
        s = S_np[:,i]
        supp = s!=0
        grad_s_supp = grad_S_np[supp,i]
        A_supp = A_np[:,supp]
        grad_X[:,i] = np.linalg.pinv(A_supp).T@grad_s_supp
    grad_X = torch.from_numpy(grad_X).double()
    return grad_X

def calc_grad_A(grad_S, A, S, X):
    """
    Calculates the gradient of q(X,A) with respect to A.

    Parameters
    ---------
    grad_S: PyTorch tensor
        The gradient of S = q(X,A) passed on from the last layer.
    A: PyTorch tensor
        The A matrix used to compute q(X,A).
    S: PyTorch tensor
        The output S = q(X,A).
    X: PyTorch tensor
        The X matrix used to compute q(X,A).

    Returns
    -------
    grad_A: PyTorch tensor
        The gradient of q(X,A) with respsect to A.

    """

    A_np = A.numpy()
    S_np = S.numpy()
    grad_S_np = grad_S.numpy()
    X_np = X.numpy()
    m = A.shape[0]
    k = A.shape[1]
    n = S.shape[1]
    grad_A = np.zeros([m,k])
    for l in range(n):
        s = S_np[:,l]
        supp = s!=0
        A_supp = A_np[:,supp]
        grad_s_supp = grad_S_np[supp,l:l+1]
        x = X_np[:,l:l+1]
        A_supp_inv = np.linalg.pinv(A_supp)
        part1 = -(A_supp_inv.T@grad_s_supp)@(x.T@A_supp_inv.T)
        part2 = (x - A_supp@(A_supp_inv@x))@((grad_s_supp.T@A_supp_inv)@A_supp_inv.T)
        grad_A[:,supp] += part1 + part2
    grad_A = torch.from_numpy(grad_A).double()
    return grad_A


class LsqNonneg(nn.Module):
    """
    Defining a submodule 'LsqNonneg' of the nn.Module with network parameter.

    Parameters
    ----------
    m: int_
        The first dimension of A, size (m,k).
    k: int_
        The second dimension of A, size (m,k).
    initial_A: PyTorch tensor, optional
        The initialization of the A matrix (default None). If None, then A is generated randomly.

    Methods
    ----------
    forward(input)
        Computes the forward propagation of the nonnegative least-squares layer.

    """
    def __init__(self, m, k, initial_A = None):
        super(LsqNonneg, self).__init__()
        self.m = m;
        self.k = k;
        self.A = nn.Parameter(torch.DoubleTensor(m,k)) #switched
        if initial_A is None:
            self.A.data = torch.abs(torch.rand(m,k,dtype = torch.double)) # initialize the network parameter #Switched
        else:
            self.A.data = initial_A
        
    def forward(self, input, last_S=None):
        """
        The forward pass of the LsqNonneg submodule.

        Parameters
        ----------
        input: PyTorch tensor
            The input the the NMF layer, X.

        Returns
        -------
        S: PyTorch tensor
            The output of the LsqNonnegF forward pass, S = q(X,A).
        """

        return LsqNonnegF.apply(input, self.A,last_S)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
    Class and functions for recording and visualizing Neural NMF results.
'''

from matplotlib import pyplot as plt


class Writer:

    """
    Records a history of the Neural NMF epochs in dictionaries and visualizes various information.

    Methods
    ----------
    add_scalar(name, scalar)
        Adds a value to the list of scalars stored under the proper key in the scalar dictionary.
    add_tensor(name, tensor)
        Adds a PyTorch tensor to the list of tensors stored under the proper key in the tensor dictionary.
    get(name)
        Fetches the list of scalars or tensors stored under name.
    plot_scalar(name)
        Plots the list of scalars stored in the scalar dictionary under the key name.
    plot_tensor(name)
        Plots the tensor stored in the idx_lst index of the list in the tensor dictionary corresponding
        to the key name.
    """

    def __init__(self):
        self.scalar_dict = {}
        self.tensor_dict = {}
        
    def add_scalar(self, name, scalar):
        """
        Adds a value to the list of scalars stored under the proper key in the scalar dictionary.

        Parameters
        ----------
        name: string
            The key value for accessing the scalar dictionary.
        scalar: float_
            The value to add to the list stored under scalar_dict[name].

        """
        if self.scalar_dict.get(name) is None:
            self.scalar_dict[name] = [scalar]
        else:
            self.scalar_dict[name].append(scalar)
            
    def add_tensor(self, name, tensor):
        """
        Adds a PyTorch tensor to the list of tensors stored under the proper key in the tensor dictionary.

        Parameters
        ----------
        name: string
            The key value for accessing the tensor dictionary.
        scalar: PyTorch tensor
            The tensor to add to the list stored under tensor_dict[name].

        """

        if self.tensor_dict.get(name) is None:
            self.tensor_dict[name] = [tensor]
        else:
            self.tensor_dict[name].append(tensor)
            
    def get(self, name):
        """
        Fetches the list of scalars or tensors stored under name. Searches the scalar list before the 
        tensor list.

        Parameters
        ----------
        name: string
            The key by which to search the scalar and tensor dictionaries. 

        Returns
        -------
        values: list, optional
            The list of scalars or tensors that correspond to name, the key value. If key value is not
            in either dictionary, returns None.

        """

        scalar =  self.scalar_dict.get(name)
        if scalar is not None:
            return scalar
        else:
            tensor = self.tensor_dict.get(name)
            if tensor is not None:
                return tensor
            else:
                print('No variable with name:', name, 'in the dictionary')
                
    def plot_scalar(self, name):
        """
        Plots the list of scalars stored in the scalar dictionary under the key name.

        Parameters
        ----------
        name: string
            The key by which to search the scalar dictionaries. 

        """

        lst = self.scalar_dict.get(name)
        plt.semilogy(lst)
        plt.show()
        
    def plot_tensor(self, name, idx_lst):
        """
        Plots the tensor stored in the idx_lst index of the list in the tensor dictionary corresponding
        to the key name.

        Parameters
        ---------
        name: string
            The key by which to search the tensor dictionaries.
        idx_list: list of int_
            The indices of the list of tensors to plot. 

        """
        tensor_lst = self.tensor_dict.get(name)
        for i in idx_lst:
            if i < len(tensor_lst):
                tensor = tensor_lst[i]
                #fig = plt.figure(figsize = (15,105))
                plt.imshow(tensor,aspect='auto',cmap='binary',interpolation='none')
                plt.show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''
    Class and functions for initializing (semi-supervised) hierarchical NMF model and forward-propagation
    of Neural NMF.
    The class consists of a sequence of factor matrices with dimensions defined by the input depth info 
    and an optional classification layer defined by the optional input number of classes.
    Examples
    --------
    >>> net = Neural_NMF([m, 9], none)
    >>> train(net, X, epoch=200, lr=1000)                            #uses function from train class
    epoch =  10 
     tensor(228.1642, dtype=torch.float64)
    epoch =  20 
     tensor(228.0765, dtype=torch.float64)
    epoch =  30 
    ...
'''

import torch
import torch.nn as nn
from torch.autograd import Variable
from matplotlib import pyplot as plt
import scipy.io as sio
from NeuralNMF import LsqNonneg
import numpy as np
import torch.nn.functional as F



class Neural_NMF(nn.Module):
    """
    Class for Neural NMF network structure.
    
    The Neural_NMF object contains several NMF layers(contained in self.lsqnonneglst, each element in 
    the list self.lsqnonneglst is a Lsqnonneg object) and a linear layer for classification(self.linear).
    Given X, the input, is mxn, this will initialize factor matrices in hierarchical NMF
        X = A_0 * A_1 * .. A_L * S_L, where:
            A_i is of size depth_info[i] x depth_info[i+1] and 
            S_L is of size depth_info[L] x n.
        If c is not None, it also initializes a classification layer defined by B*S_L where:
            B is of size c x depth_info[L].
    ...
    Parameters
    ----------
    depth_info: list
        The list [m, k1, k2,...k_L] contains the dimension information for all factor matrices.
    c: int_, optional
        Number of classes (default is None).
    Methods
    ----------
    forward(X)
        Forward propagate the Neural NMF network.
    """
    def __init__(self, depth_info, c = None):
        super(Neural_NMF, self).__init__()
        self.depth_info = depth_info
        self.depth = len(depth_info)
        self.c = c
        self.lsqnonneglst = nn.ModuleList([LsqNonneg(depth_info[i], depth_info[i+1]) 
                                           for i in range(self.depth-1)]) 
                                           #ititalized a list of Nonnegative least squared objects
        if c is not None:
            self.linear = nn.Linear(depth_info[-1],c, bias = False).double() 
                                        #initialize classification layer (with last factor matrix)
            with torch.no_grad():
                self.linear.weight.copy_(torch.abs(torch.rand(depth_info[-1],c)))


    def forward(self, X, Y=None, L=None, last_S_lst=None):
        """
        Runs the forward pass of the Neural NMF network.
        Parameters
        ----------
        X: PyTorch tensor
            The m x n input to the Neural NMF network. The first dimension, m, should match the first entry 
            of depth_info.
        Returns
        -------
        S_lst: list
            All S matrices ([S_0, S_1, ..., S_L]) calculated by the forward pass.
        pred: PyTorch tensor, optional
            The c x n output of the linear classification layer.
        """

        if len(X.shape) != 2:
            raise ValueError("Expected a two-dimensional Tensor, but X is of shape {}".format(X.shape))

        if X.shape[0] != self.depth_info[0]:
            raise ValueError("Dimension 0 of X should match entry 0 of depth_info, but values were {} and {}".format(X.shape[0], self.depth_info[0]))

        S_lst = []
        for i in range(self.depth-1):
            if last_S_lst == None:
                X = self.lsqnonneglst[i](X)  #Calculates the least squares objective S = min S>=0 ||X - AS||
            else:
                X = self.lsqnonneglst[i](X, last_S_lst[i]) #Calculates the least squares objective S = min S>=0 ||X - AS||
            S_lst.append(X)
        #once at end of network (X = S^L)
        if self.c is None:
            return S_lst
        else: #if in supervised case, multiply by B to produce pred = B*S^L

            #B = torch.mm(Y[:,L[0]==1], torch.pinverse(X[:,L[0]==1]))
            
            #B = torch.mm(Y, torch.pinverse(X))
            #pred = torch.transpose(self.linear(torch.transpose(X,0,1)),0,1)
            #pred = torch.mm(B,X)
            pred = self.linear(X) #Jamie testing
            return S_lst, pred
        





"""Top-level package for Neural NMF."""

# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "0.0.10"

__author__ = "Joshua Vendrow"
__email__ = "jvendrow@ucla.edu"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2020, Joshua Vendrow"

from .writer import *
from .news_group_loading import *

from .loss import *
from .lsqnonneg_module import *
from .model import *
from .train import *



from setuptools import setup, find_packages


setup(
    name='NeuralNMF',
    version='0.0.10',
    license='MIT',
    author="Joshua Vendrow",
    author_email='jvendrow@ucla.edu',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    url='https://github.com/jvendrow/NeuralNMF',
    keywords='neural nmf',
    install_requires=[
          'torch',
          'matplotlib',
          'scipy',
          'numpy',
          'fnnls',
          'tqdm',
      ],

)